/********************************************************************/
/*                                                                  */
/*   Program Name :  speana_n(ew)                                   */
/*                                                                  */
/*      Version:  1.00  2015-11-04                                  */
/*                                                                  */
/*      Copyright (c) 2015-     T.Kondo All Right Reserved          */
/*                                                                  */
/********************************************************************/
/*

    renew speana.c

          T.Kondo 2002-05-20
               Ver 0.00 2015-11-02 start development based on speana.c 
               Ver 1.00 2015-11-04 
               Ver 1.10 2015-12-22 

    How to run:

       speana_n filename [options]

        where filename -- sampling data file name

       [options]
                    -m[ode] mode
                        set axis mode or autocorrelation mode
                            mode=0 : amplitude log, frequency log
                                =1 (deflt) : amplitude log, frequency linear
                                =10 : amplitude linear, frequency log
                                =11 : amplitude linear, frequency linear
                                =-1 : Auto correlation function plot
                                      instead of spectrum
                                =-N : N points auto correlation function plot
                    -p[mode] pmode
                         plot device selection
                            pmode=0 :  display and PostScript file out
                                 =1 :  PostScript file out only
                                 =2 :  display only
                                default is 2
                    -a[verage]|-i[ntegration]|-t[span] sekibun
                         set averaging period (sec)
                                 default is 0.01
                    -c[omment] comment
                         comment appeared on a graph
                    -o[ffset] soffset
                         set start time offset in sec
                                 default is 0.0 
                    -f1[khz] f1khz
                         set display frequency range (lower boundary) (kHz)
                                 for power Maximum (default is 0.0)
                    -f2[khz] f2khz
                         set display frequency range (upper boundary) (kHz)
                                 for power Maximum (default is video frequency)
                    -min[dbm] mindbm
                         set display amplitude minimum (dBm) (default auto)
                    -max[dbm] maxdbm
                         set display amplitude maximum (dBm) (default auto)
                    -nops  --- same as "-p 2" (no postscript out)  <== no more used!!
                    -h[alt]  --- halt after each plot (default continuous plot)
                    -s[msec] smsec
                         set suspend time (msec) after each plot
                         default is 0 (as soon as posiible base)
             
*/
#define PROGRAM_VERSION "Ver. 2015-12-22"

//#define USE_GNUPLOT    // use GNUPLOT  (default is use of PGPLOT)

#include "speana_n.h"    // header for graphics
#include "g_head.h"    // header for graphics

#include "libk5.h"      // common header for IP-VLBI utilities

#include "libfft.h"

#ifdef WIN32
#define USE_ASM    // use inline assemble function
#endif

#include "libvssp32.h"
#include "libspeplot.h"

/*    Offset from dB to dBm conversion from A/D multibit sampling data */ 
#define OFFSET1  16.0     //1 bit sampling full power convert to 0 dB (theoretical)
      /* following constants are provisional  (measured on 2002/5/18) */
#define OFFSET2   9.0     //2 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET4  -1.0     //4 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET8 -26.0     //8 bit sampling data offset measured at 200kHz for PC(vlbi11)
#define OFFSET16 -78.0     //16 bit sampling data offset (this is for USBADC adjusting value) 2015-10-27

#define DEFLT_DATA_FILE "tds.data"   // default data file name

//#define DEFLT_SLEEP_MSEC 100        // default sleep suspend time in msec

//#ifdef WIN32
//int tempcounter;  // counter for tempfile manage
//#endif

void prepare_plotdata(speana_param_t *st, fftp_t *gp, int *scount);

/* global variables */
fftp_t gp;      // for FFT use



void print_usage1(speana_param_t *st)
{
    char pgdevice_dflt[20];
    printf("speana_n %s\n",PROGRAM_VERSION);
#ifdef USE_FFTW
    printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
    printf(" compliled for GNUPLOT \n");
    strcpy(pgdevice_dflt,G_DISPLAY);
#else
    printf(" compliled for PGPLOT \n");
    strcpy(pgdevice_dflt,st->g_display);
#endif
    printf("\n");
    printf("Usage: speana_n filename [options]\n");
    printf("\n");
    printf("         where  filename ---- Data file name\n");
    printf("\n");
    printf("            [options]\n");
    printf("                -m[ode] mode\n");
    printf("                    set axis mode or autocorrelation mode\n");
    printf("                        mode=0 : amplitude log, frequency log\n");
    printf("                            =1 (deflt) : amplitude log, frequency linear\n");
    printf("                            =10 : amplitude linear, frequency log\n");
    printf("                            =11 : amplitude linear, frequency linear\n");
    printf("                            =-1 : Auto correlation function plot\n");
    printf("                                  instead of spectrum\n");
    printf("                            =-N : N points auto correlation function plot\n");
//    printf("                -p[mode] pmode\n");
//    printf("                     plot device selection\n");
//    printf("                        pmode=0 :  display and PostScript file (%s) out\n",st->ps_fname);
//    printf("                             =1 :  PostScript file (%s) out only\n",st->ps_fname);
//    printf("                             =2 :  display only\n");
//    printf("                            default is 2\n");
    printf("                -a[verage]|-i[ntegration]|-t[span] sekibun\n");
    printf("                     set averaging period (sec)\n");
    printf("                             default is 0.01\n");
//    printf("                -c[omment] comment\n");
//    printf("                     comment appeared on a graph\n");
    printf("                -o[ffset] soffset\n");
    printf("                     set start time offset in sec\n");
    printf("                             default is 0.0\n"); 
    printf("                -f1[khz] f1khz\n");
    printf("                     set display frequency range (lower boundary) (kHz)\n");
    printf("                             for power Maximum (default is 0.0)\n");
    printf("                -f2[khz] f2khz\n");
    printf("                     set display frequency range (upper boundary) (kHz)\n");
    printf("                             for power Maximum (default is video frequency)\n");
    printf("                -min[dbm] mindbm\n");
    printf("                     set display amplitude minimum (dBm) (default auto)\n");
    printf("                -max[dbm] maxdbm\n");
    printf("                     set display amplitude maximum (dBm) (default auto)\n");
    //printf("                -nops  --- same as \"-p 2\" (no postscript out)\n");
    printf("                -h[alt]  --- halt after each plot (default continuous plot)\n");
    printf("                -s[msec] smsec\n");
    printf("                     set suspend time (msec) after each plot\n");
    printf("                     default is 0 (as soon as possible base)\n");
    printf("\n");
#ifndef USE_GNUPLOT   // using gnuplot
    printf("\n");
    printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
    printf("                   = %s",getenv(PGDISP));
    if(getenv(PGDISP)==NULL) {
        printf(",  program default is ( %s )\n",PGDISP_DFLT);
    } else {
        printf("\n");
    }
#endif
    printf("\n");
    exit(0);
}

void runparam_get1(int argc, char *argv[],runp_t *runp)                       
{
    /* get run parameters type 1 (conventional run tyep) */

    //int n,i;
    char cbuf[256];

	// initialize parameters

    runp->fname[0]='\0';
    runp->comment[0]='\0';
    runp->soffset=0.0;
    //runp->sekibun=1.0;
    runp->sekibun=0.01;
    runp->mode2=1;
    runp->mode=1;  // default frequency axis mode (=Linear scale)
    runp->ymode=0; // default amplitude axis mode (=LOG scale)
    runp->amode=0; // auto correlation mode off
    runp->pmode=2; // default plot mode
    runp->f1khz=0.0;
    runp->f2khz=0.0;
    runp->mindbm=0.0;
    runp->maxdbm=0.0;
    runp->flag_halt=0;
    runp->sleep_msec=0;

    if (argc>=2){
        strcpy(runp->fname,*(argv+1));
    }
    if (strlen(*(argv+1))==1 && runp->fname[0]=='0'){
          strcpy(runp->fname, DEFLT_DATA_FILE);
    }
 

    if(cmnd_get(argc,argv,"-m",cbuf)==0){                 // -mode 
       runp->mode2 = atoi(cbuf);  
       if(runp->mode2 <0 ){
            runp->amode=-runp->mode2;
            runp->mode=0;
       } else {
            runp->amode=0;
            runp->ymode=runp->mode2/10;
            runp->mode=runp->mode2 % 10;
       }
    }
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->pmode = atoi(cbuf);     // -pmode 
    if(cmnd_get(argc,argv,"-c",cbuf)==0) strcpy(runp->comment,cbuf);   // -comment 
    if(cmnd_get(argc,argv,"-i",cbuf)==0) runp->sekibun = atof(cbuf);   // -integration 
    if(cmnd_get(argc,argv,"-a",cbuf)==0) runp->sekibun = atof(cbuf);   // -average 
    if(cmnd_get(argc,argv,"-t",cbuf)==0) runp->sekibun = atof(cbuf);   // -tspan 
    if(cmnd_get(argc,argv,"-o",cbuf)==0) runp->soffset = atof(cbuf);   //  -offset 
    if(cmnd_get(argc,argv,"-f1",cbuf)==0) runp->f1khz = atof(cbuf);    //  -f1khz 
    if(cmnd_get(argc,argv,"-f2",cbuf)==0) runp->f2khz = atof(cbuf);    //  -f2khz 
    if(cmnd_get(argc,argv,"-min",cbuf)==0) runp->mindbm = atof(cbuf);  //  -mindbm 
    if(cmnd_get(argc,argv,"-max",cbuf)==0) runp->maxdbm = atof(cbuf);  //  -maxdbm 
    if(cmnd_get(argc,argv,"-nops",cbuf)==0) runp->pmode=2;             //  -nops
    if(cmnd_get(argc,argv,"-h",cbuf)==0) runp->flag_halt=1;            // -halt
    if(cmnd_get(argc,argv,"-s",cbuf)==0) runp->sleep_msec = atoi(cbuf);   //  -sleep_msec 

}

void monit_param(runp_t *runp)
{
    printf("  fname   : %s\n",runp->fname);
    printf("  mode2   : %d\n",runp->mode2);
    printf("  mode    : %d\n",runp->mode);
    printf("  amode   : %d\n",runp->amode);
    printf("  ymode   : %d\n",runp->ymode);
    printf("  sekibun : %f\n",runp->sekibun);
    printf("  soffset : %f\n",runp->soffset);
    printf("  f1khz   : %f\n",runp->f1khz);
    printf("  f2khz   : %f\n",runp->f2khz);
    printf("  mindbm  : %f\n",runp->mindbm);
    printf("  maxdbm  : %f\n",runp->maxdbm);
    printf("  sleep_msec : %d\n",runp->sleep_msec);
}


int main(int argc, char *argv[])
{
    //int nloop,sekibun;
    char cw;
    int nloop;
    int flag_last;   // 2015-10-27
    int i,imax, m, k, kaisu;
    int smode,usampl,nfft;
	int kbytes;
    int64_t bits, bytes;  // 2105-10-23
    int64_t ns_bopp;      // sample# from the 1st point of data file at begining of tsapn period
    int64_t ns_all;       // sample# from the 1st point of data file
    int numb;
    //int amode;    // >0 auto correlation plot mode 
    //int mode;  // frequency axis mode  0: log  1: linear 
    //int mode2;  // 2 digits axis mode
    int ymode; // amplitude scale mode   0: log   1: linear
    int scount[4];
    int p_usampl;     // current sample # in USAMPL
    int p_adata;      // current sample # in ADATA
    int flag_plot_done; 
    float fsampl,tsampl,fact;
    //double *cdata; => global
    //double *f_hanning;
    double theta;
    double sum,dc;
    float pows;
    //float ymaxdb,ymindb;
//#ifdef USE_FFTW
//    fftw_complex *cc;  //=> global
//    fftw_plan pfcc, pbcc;  // => global
//#endif
    int isoffset;     // start offset in integer second      2014-08-12
    double fsoffset;  // fractional part of start offset     2014-08-12
    double ftime;     // time in second to judge start time  2014-08-12
    int start_flag;   // flag to judge start time            2014-08-12
    double nfact;
    double tinteg,tusampl;
    double telapsed;           // to measure processing time
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char ltime1st[60];
    char cbuf[20];
    //char pgdevice_dflt[20];
    //char* dflt_fname="tds.data";
    runp_t runp;
    K5VSSP_file_t K5;
    speana_param_t st;

    /* set graphic device */
    strcpy(st.ps_fname,PS_OUT_FILE);
#ifdef USE_GNUPLOT   // using gnuplot
    strcpy(st.g_display,G_DISPLAY);
    //strcpy(pgdevice_dflt,G_DISPLAY);
    strcpy(st.g_ps_device,G_PS_DEVICE);
#else
    strcpy(st.g_display,select_env_dflt(PGDISP, PGDISP_DFLT));
    //strcpy(pgdevice_dflt,st.g_display);
    sprintf(st.g_ps_device,"%s%s",st.ps_fname,G_PS_DEVICE);
#endif

    if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){
         print_usage1(&st);
         //printf(" Hit return key for another how to run ");
         //cw=getchar();  /* dummy read to accept just return */
         //print_usage2();
         exit(0);
    }
    if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ 
          printf("speana %s",PROGRAM_VERSION);
#ifdef USE_FFTW
          printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
          printf(" compliled for GNUPLOT \n");
#else
          printf(" compliled for PGPLOT \n");
#endif
          printf("\n");
          exit(0);
    }
    
    /* start processing */

     printf("speana %s",PROGRAM_VERSION);
     gp.flag_fftw=0;
#ifdef USE_FFTW
     gp.flag_fftw=1;
     printf(" compliled with FFTW3.0 \n");
#endif
#ifdef USE_GNUPLOT   /* using gnuplot */
     printf(" compliled for GNUPLOT \n");
#else
     printf(" compliled for PGPLOT \n");
#endif
     printf("\n");

    /* get run parameters */
    runparam_get1(argc, argv, &runp);

    //monit_param(&runp); exit(0);  // debug

    /* copy runparameters to other variables */
    strcpy(st.fname,runp.fname);
    strcpy(K5.fname,st.fname);

    st.mode2=runp.mode2;
    st.mode=runp.mode;
    st.amode=runp.amode;
    ymode=runp.ymode;
    st.sekibun=runp.sekibun;

    st.pmode=runp.pmode;
    strcpy(st.comment,runp.comment);
    st.soffset=runp.soffset;
    st.toffset=runp.soffset;

    isoffset=(int)st.soffset;
    fsoffset=fmod(st.soffset,1.0);
    if(fsoffset > 0.0){
        start_flag=0;
    } else { 
        start_flag=1;
    }
    ftime=0.0;

    st.f1khz=runp.f1khz;
    st.f2khz=runp.f2khz;
    st.mindbm=runp.mindbm;
    st.maxdbm=runp.maxdbm;

    st.flag_halt=runp.flag_halt;
    st.sleep_msec=runp.sleep_msec;

    cstart1sec = clock();

    printf("Data File is %s\n",K5.fname);
    K5.size=get_file_size(K5.fname);
    if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",K5.fname);
        goto exitend2;
    }
    k5vssp_t_ini(&K5);

    /* Data header check to get sampling parameters */

    if((kbytes=CheckK5header_with_Read(&K5)) < 0){
        printf("File read error! \n");
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    /* parameter set */

    flag_last=0;   // 2015-10-27

    st.sfreq=K5.sfreq;
    st.adbit=K5.adbit;
    st.numch=K5.numch;
    fsampl=(float)(1000.0*(float)st.sfreq);   // sampling frequency (Hz)
    tsampl=(float)(1.0/fsampl);            // sampling period (sec)
    //bits=((long)fsampl)*st.adbit*st.numch;  // # of bits in 1 sec data except header
    //bits=((int)fsampl)*st.adbit*st.numch;  // # of bits in 1 sec data except header
    bits=((int64_t)fsampl)*st.adbit*st.numch;  // # of bits in 1 sec data except header  2015-10-23
    bytes=bits/8;                     // # of bytes in 1 sec data except header
    K5.fsample=fsampl;
    //sekibun=1;                       // Averaging period for Spectrum (sec)
    K5.bytesec=(int)(bits/8);                  //# of bytes in 1 sec data

    smode=1;                         // low resolution mode

    if(smode ==1){ 
        usampl=10000; 
    } else {
        usampl=50000;
        if(st.sfreq == 40)  usampl=40000; 
        if(st.sfreq == 100) usampl=100000; 
        if(st.sfreq == 200) usampl=200000; 
    }
    imax=(st.sfreq*1000)/usampl;
    numb=usampl*st.adbit*st.numch/8;   // # of bytes in unit # of samples
    K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
    K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second


    tusampl=(double)usampl/K5.fsample;   // period of usampl

    printf("imax numb  %d %d\n",K5.imax,K5.numb);    //debug

    nfft=1024;
    if(usampl < 524288) nfft= 524288;
    if(usampl < 262144) nfft= 262144;
    if(usampl < 131072) nfft= 131072;
    if(usampl < 65536)  nfft= 65536;
    if(usampl < 32768)  nfft= 32768;
    if(usampl < 16384)  nfft= 16384;

    st.nfft=nfft;
    gp.nfft=nfft;

    nfact=2.0/(double)nfft;

    K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl
    K5.tsample=1.0/K5.fsample;    // 2015-10-23
    K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

    K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
    K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

    #ifdef USE_FFTW
        gp.cc=fftw_malloc(sizeof(fftw_complex)*gp.nfft);

        gp.pfcc=fftw_plan_dft_1d(gp.nfft, gp.cc, gp.cc, FFTW_FORWARD, FFTW_ESTIMATE);
        gp.pbcc=fftw_plan_dft_1d(gp.nfft, gp.cc, gp.cc, FFTW_BACKWARD, FFTW_ESTIMATE);
    #else
        gp.cdata = (double *)malloc(2*gp.nfft*sizeof(double));
    #endif
    st.spows = (float *)malloc(4*gp.nfft*sizeof(float));
    gp.f_hanning = (double *)malloc(gp.nfft*sizeof(double));

    if(st.amode==0){ // spectrum mode
        //Hanning filter coeeficients prepare
        for(k=0; k<gp.nfft; k++){
           theta=360.0*(double)k/(double)(gp.nfft-1) - 180.0;
           theta=fmod(theta,360.0) ;
           gp.f_hanning[k]=0.5*(1.0+cos(theta*PIRA));
        }
    } else {         // auto correlation mode  
        for(k=0; k<gp.nfft; k++) gp.f_hanning[k]=1.0;   
    }

    fact=((float)usampl)*((float)usampl);

    /* foward startffset seconds */
    //if(fwdNsecK5(st.soffset, &K5) <0){
    if(fwdNsecK5(isoffset, &K5) <0){            // 2014-08-12
        //printf("Foward error %d\n",st.soffset);
        printf("Foward error %d\n",isoffset);
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    for(k=0; k<4*nfft; k++) st.spows[k]=0.0;
    for(k=0; k<4; k++) scount[k]=0;

    printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
        K5.hh,K5.mm,K5.ss,K5.seconds);
    if(K5.format==2){  // VSSP32          2014-06/13
	   sprintf(ltime1st,"Time(UT) %04d/%03d %02d:%02d:%02d",
           K5.year,K5.totalday,K5.hh,K5.mm,K5.ss);
    } else {
       sprintf(ltime1st,"Time(UT) %02d:%02d:%02d",K5.hh,K5.mm,K5.ss);
    }
    if(fsoffset > 0.0){   // 2014-08-12
        sprintf(cbuf,".%03d",(int)(1000.0*fsoffset));
        strcat(ltime1st,cbuf);
    }
    strcpy(st.lab2,ltime1st);
    strcat(st.lab2,"  ");
    strcat(st.lab2,st.comment);
    
    ns_all=0;
    ns_bopp=0;
        
    tinteg=0.0;

    st.flag_g_device_open=0;

    nloop=0;
loop1:

    /* 1sec loop entry point */

    if(nloop >0) {
        if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            i=fclose( K5.stream ) ;
            goto exitend;
        }
        printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
            K5.hh,K5.mm,K5.ss,K5.seconds);
    }

    /* usampl read entry */

    for(i=0; i<K5.imax; i++){
        K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
        if(K5.wordread < K5.nword){  // probably no more data     2015-10-27
            flag_last=1;
            if(K5.wordread==0){
                tusampl=0;
                goto exithere;
            }
            /* recalc tusampl and fact for this reading */
            if(st.adbit*st.numch >= 8){
                usampl=K5.wordread*4/(st.adbit*st.numch/8);
                fact=((float)usampl)*((float)usampl);
                tusampl=(double)usampl/K5.fsample;   // period of usampl
            } else {
                tusampl=0;
                goto exithere;
            }
        }

        K5.pos+=K5.wordread*4;
        K5.vflag=1; // validiate data
        parseADdata_3N(&K5);

        if(start_flag==0){
            if(ftime < fsoffset){
                ftime+=tusampl;
                ns_all+=(K5.npos/st.numch);
                continue;
            } else {
                ns_bopp=ns_all;
                start_flag=1;
            }
        }


        for(m=0; m<st.numch; m++){
            kaisu=0;
            sum=0.0;

            // initialize array
            #ifdef USE_FFTW
                for(k=0; k<gp.nfft; k++){ 
                    gp.cc[k][0]=0.0; gp.cc[k][1]=0.0;
                }
            #else
                for(k=0; k<2*gp.nfft; k++) gp.cdata[k]=0.0;
            #endif

            for(k=0; k<K5.npos/K5.numch; k++){

                /* set data */
                #ifdef USE_FFTW
                   gp.cc[k][0]=(double)K5.idata[k*st.numch+m];  
                   sum+=gp.cc[k][0];
                   //printf(" %f",(double)K5.idata[k*st.numch+m]); // debug
                #else
                   gp.cdata[2*k]=(double)K5.idata[k*st.numch+m];  
                   sum+=gp.cdata[2*k];
                #endif
                kaisu++;
            }
             
            dc=sum/(double)kaisu;    // DC component
     
            // DC offset removal from data
            for(k=0; k<K5.npos/st.numch; k++){
                #ifdef USE_FFTW
                   gp.cc[k][0]-=dc;
                #else
                   gp.cdata[2*k]-=dc;
                #endif
            }

            /* FFT (time -> frequency ) */

            for(k=0; k<gp.nfft; k++){
                #ifdef USE_FFTW
                   gp.cc[k][0]*=gp.f_hanning[k];  // hanning filter
                   gp.cc[k][1]*=gp.f_hanning[k];  // hanning filter
                #else
                   gp.cdata[2*k]*=gp.f_hanning[k];
                   gp.cdata[2*k+1]*=gp.f_hanning[k];
                #endif
            }

            #ifdef USE_FFTW
               //fftw_execute(pbcc);
               fftw_execute(gp.pfcc);    // 2015-11-03
            #else
               //cdft(2*nfft,1,cdata);
               cdft(2*gp.nfft,-1,gp.cdata);   // 2015-11-03
               for(k=0; k<2*gp.nfft; k++){
                   gp.cdata[k]*=(double)nfft;    // this correction is due to cdft() used here
               }
            #endif

            /* calculate power  */ 
            for(k=0; k<gp.nfft; k++){
               #ifdef USE_FFTW
                   pows=(float)((gp.cc[k][0]*gp.cc[k][0]+gp.cc[k][1]*gp.cc[k][1])/(double)fact);
               #else
                   pows=(float)((gp.cdata[2*k]*gp.cdata[2*k]+gp.cdata[2*k+1]*gp.cdata[2*k+1])/(double)fact);
               #endif
                st.spows[k*4+m]=st.spows[k*4+m]+pows;
            }

            scount[m]++;
        }
exithere:

        tinteg+=tusampl;
        //st.toffset=(double)ns_bopp*tsampl;
        if(flag_last == 1){  // 2015-10-27
            if(tinteg==0.0) goto exitloop;
            if(st.sekibun > tinteg){
               st.sekibun=tinteg;
            }
            prepare_plotdata(&st, &gp, scount);
            st.toffset=(double)ns_bopp*tsampl;
            plotspe_new(&st);
            ns_bopp=ns_all;
            //st.toffset+=tusampl;
            for(k=0; k<4*gp.nfft; k++) st.spows[k]=0.0;
            for(k=0; k<4; k++) scount[k]=0;
            goto exitloop;  
        }
        if(tinteg >= st.sekibun){
             prepare_plotdata(&st, &gp, scount);
             st.toffset=(double)ns_bopp*tsampl;
             plotspe_new(&st);
             ns_bopp=ns_all;
             //st.toffset+=tusampl;
             for(k=0; k<4*gp.nfft; k++) st.spows[k]=0.0;
             for(k=0; k<4; k++) scount[k]=0;
             tinteg=0.0;
             if(st.flag_halt==1){
                //printf(" Hit return key for next span (CTRL-C for stop) ");
                printf(" Hit return key for next span (or \"N\"on-stop or \"Q\"uit) ");
                cw=getchar();  /* dummy read to accept just return */
                if(cw=='N' || cw=='n') st.flag_halt=0;
                if(cw=='Q' || cw=='q') goto exitloop;
            } else {
                 if(st.sleep_msec >= 0) {
                     spe_sleep(st.sleep_msec);
                 } else {
                     goto exitloop;
                 }
             }
            //goto exitloop;
        }
        ns_all+=(K5.npos/st.numch);
        //st.toffset+=tusampl;

    }
    goto loop1;
exitloop:
    i=fclose( K5.stream ) ;

    //prepare_plotdata(&st, &gp, scount);

    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);

    /* plot spectrum (or auto cor) POSTSCRIPT OUT and XTERM*/
   //  plotspe(&st);

  exitend: ;
      
    free(K5.ubuf);
    free(K5.idata);
    free(st.spows);
    free(gp.f_hanning);

    #ifdef USE_FFTW
       fftw_destroy_plan(gp.pfcc);
       fftw_destroy_plan(gp.pbcc);
       fftw_free(gp.cc);
    #else
       free(gp.cdata);
    #endif

 exitend2: ;

    return 0;
}


void prepare_plotdata(speana_param_t *st, fftp_t *gp, int *scount)
{
    int m,k;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;

    if(st->amode ==1) {   // auto correlation function calc
       for(m=0; m<st->numch; m++){
          for(k=0; k<gp->nfft; k++){
             st->spows[k*4+m]/=scount[m]; // nomalization
             #ifdef USE_FFTW
                gp->cc[k][0]=st->spows[k*4+m]; gp->cc[k][1]=0.0;
             #else
                gp->cdata[2*k]=st->spows[k*4+m];
                gp->cdata[2*k+1]=0.0;
             #endif
          }
          if(m==1 || m==3){
             #ifdef USE_FFTW
                gp->cc[0][0]=1.0;
             #else
                gp->cdata[0]=1.0;  // to avoid math error
             #endif
          }
          /* FFT (frequency -> time) */
          #ifdef USE_FFTW
             fftw_execute(gp->pbcc);
          #else
             cdft(2*gp->nfft,1,gp->cdata);
          #endif

          for(k=0; k<gp->nfft; k++){
             #ifdef USE_FFTW
                if(k==0) pmax=(float)gp->cc[0][0];
                if(gp->cc[k][0] > pmax) pmax=(float)gp->cc[k][0];
                st->spows[k*4+m]=(float)gp->cc[k][0];
             #else
                if(k==0) pmax=(float)gp->cdata[0];
                if(gp->cdata[2*k] > pmax) pmax=(float)gp->cdata[2*k];
                st->spows[k*4+m]=(float)gp->cdata[2*k];
             #endif
          }
          for(k=0; k<gp->nfft; k++){  // normalize
             st->spows[k*4+m]/=pmax;
          }
             
       }
       for(k=0; k<4; k++) scount[k]=0;  // scount reset
         
    } else {  // power spectrum

        /* convert power to dB  then dBm */       
        cofst=OFFSET1;       //1 bit sampling full power convert to 0 dB
        if( st->adbit==8) cofst=OFFSET8;     // 8bit sampling offset
        if( st->adbit==4) cofst=OFFSET4;     // 4bit sampling offset
        if( st->adbit==2) cofst=OFFSET2;     // 2bit sampling offset
        if( st->adbit==16) cofst=OFFSET16;     // 16bit sampling offset (arbitrary value) added on 2015-10-23
        for(m=0; m<st->numch; m++){
            for(k=0; k<gp->nfft; k++){
                st->spows[k*4+m]/=scount[m]; // nomalization
                st->spows[k*4+m]=(float)(10.0*log10(st->spows[k*4+m]) + cofst); 
                if(k==0){
                    pmaxdbm=st->spows[k*4+m];
                    pmindbm=st->spows[k*4+m];
                    if(m==0){
                        ymaxdb=pmaxdbm;
                        ymindb=pmindbm;
                    }
                } else {
                    if(st->spows[k*4+m] > pmaxdbm) pmaxdbm=st->spows[k*4+m]; 
                    if(st->spows[k*4+m] < pmindbm) pmindbm=st->spows[k*4+m]; 
                }
            }
            pmax=pmaxdbm-cofst;
            pmin=pmindbm-cofst;
//            printf("CH# %d: Maximum data (dB, dBm) is %f %f\n", m+1,pmax,pmaxdbm);      
            if(pmaxdbm > ymaxdb) ymaxdb=pmaxdbm;
            if(pmindbm < ymindb) ymindb=pmindbm;
        }
        st->ymaxdb=ymaxdb;
        st->ymindb=ymindb;
//        printf("All CH: Maximum and Minimum data (dBm) are %f %f\n", ymaxdb, ymindb);      

    }
}
